# -*- coding: utf-8 -*-
"""
The base convergence WorkChain
"""
from abc import ABCMeta, abstractmethod
import typing as ty
import pathlib
import yaml

from aiida_tools.process_inputs import get_fullname

from aiida.engine import WorkChain, ToContext
from aiida import orm
from aiida.plugins import WorkflowFactory, DataFactory

from aiida_sssp_workflow.workflows.convergence.engine import TwoInputsTwoFactorsConvergence
from aiida_sssp_workflow.utils import helper_parse_upf
from aiida_sssp_workflow.helpers import get_pw_inputs_from_pseudo

CreateEvaluateWorkChain = WorkflowFactory('optimize.wrappers.create_evaluate')
OptimizationWorkChain = WorkflowFactory('optimize.optimize')
UpfData = DataFactory('pseudo.upf')

PARA_ECUTWFC_LIST = lambda: orm.List(list=[
    20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100, 110,
    120, 130, 140, 160, 180, 200
])

PARA_ECUTRHO_LIST = lambda: orm.List(list=[
    160, 200, 240, 280, 320, 360, 400, 440, 480, 520, 560, 600, 640, 680, 720,
    760, 800, 880, 960, 1040, 1120, 1280, 1440, 1600
])


class BaseConvergenceWorkChain(WorkChain):
    """Meta WorkChain to run convergence test"""
    __metaclass__ = ABCMeta

    # hard code parameters of convergence workflow
    # can be(should be) overwrite by subclass
    _TOLERANCE = 1e-1
    _CONV_THR_CONV = 1e-1
    _CONV_WINDOW = 3

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        # yapf: disable
        super().define(spec)
        spec.input('pseudo', valid_type=UpfData, required=True,
                   help='Pseudopotential to be verified')
        spec.input('options', valid_type=orm.Dict, required=False,
                   help='Optional `options` to use for the `PwCalculations`.')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                   help='The protocol to use for the workchain.')
        spec.input_namespace('parameters', help='Para')
        spec.input('parameters.ecutrho_list', valid_type=orm.List, default=PARA_ECUTRHO_LIST,
                   help='dual value for ecutrho list.')
        spec.input('parameters.ecutwfc_list', valid_type=orm.List, default=PARA_ECUTWFC_LIST,
                   help='list of ecutwfc evaluate list.')
        spec.input('parameters.ref_cutoff_pair', valid_type=orm.List, required=True,
                   default=lambda: orm.List(list=[200, 1600]),
                   help='ecutwfc/ecutrho pair for reference calculation.')
        spec.outline(
            cls.setup_protocol,
            cls.init_step,
            cls.setup,
            cls.validate_structure,
            cls.run_ref,
            cls.run_all,
            cls.results,
            cls.final_step,
        )
        spec.output('output_parameters', valid_type=orm.Dict, required=True,
                    help='The output parameters include results of all calculations.')
        spec.output('xy_data_ecutwfc', valid_type=orm.XyData, required=True,
                    help='The output XY data for plot use; the x axis is ecutwfc.')
        spec.output('xy_data_ecutrho', valid_type=orm.XyData, required=True,
                    help='The output XY data for plot use; the x axis is ecutrho.')
        spec.output('output_convergence_parameters', valid_type=orm.Dict, required=False,
                    help='The result point of convergence test.')
        spec.output('output_pseudo_header', valid_type=orm.Dict, required=True,
                    help='The header(important parameters) of the pseudopotential.')
        spec.exit_code(400, 'ERROR_SUB_PROCESS_FAILED',
                    message='The sub processes {pk} did not finish successfully.')
        spec.exit_code(600, 'ERROR_NOT_ENOUGH_EVALUATE_WORKFLOW',
                    message='The number of sub evaluation processes {n} is not enough.')
        spec.exit_code(510, 'ERROR_DIFFERENT_SIZE_OF_ECUTOFF_PAIRS',
                    message='The ecutwfc_list and ecutrho_list have incompatible size.')
        # yapf: enable

    def _get_protocol(self):
        """Load and read protocol from faml file to a verbose dict"""
        with open(
                str(
                    pathlib.Path(__file__).resolve().parents[1] /
                    'protocol.yml')) as handle:
            self._protocol = yaml.safe_load(handle)  # pylint: disable=attribute-defined-outside-init

            return self._protocol

    @abstractmethod
    def setup_protocol(self):
        """
        For different convergence sub-workflow implement only needed parameters
        from protocol.
        """

    @abstractmethod
    def get_create_process(self):
        """
        the create process
        """

    @abstractmethod
    def get_evaluate_process(self):
        """
        the evaluate process
        """

    @abstractmethod
    def get_parsed_results(self) -> ty.Dict[str, str]:
        """
        Defined the return values will be recorded in output_parameters
        This method return a dict with keys as name of results and values as
        the unit of corresponding results.
        The results are read from the evaluate helper function defined in `get_evaluate_process`.
        """

    @abstractmethod
    def get_converge_y(self) -> ty.Tuple[str, ty.Tuple[str, str]]:
        """
        The name of value in the output of evaluate process, use in as the
        convergence value.
        The first element of tuple is the name of the value and the second is
        a tuple of (<description>, <unit>).
        """

    def init_step(self):
        """
        A empty initial step which customized by user
        """

    def final_step(self):
        """
        A empty final step which customized by user
        """

    def setup(self):
        """setup"""
        self.ctx.ecutwfc_list = self.inputs.parameters.ecutwfc_list.get_list()
        self.ctx.ecutrho_list = self.inputs.parameters.ecutrho_list.get_list()
        if not len(self.ctx.ecutwfc_list) == len(self.ctx.ecutrho_list):
            return self.exit_codes.ERROR_DIFFERENT_SIZE_OF_ECUTOFF_PAIRS

        # parse pseudo and output its header information
        upf_info = helper_parse_upf(self.inputs.pseudo)
        self.ctx.element = self.inputs.pseudo.element

        self.out('output_pseudo_header', orm.Dict(dict=upf_info).store())

    def validate_structure(self):
        """validate structure"""
        res = get_pw_inputs_from_pseudo(pseudo=self.inputs.pseudo)

        self.ctx.structure = res['structure']
        self.ctx.pseudos = res['pseudos']
        self.ctx.base_pw_parameters = res['base_pw_parameters']

    def run_ref(self):
        """
        Running the calculation for the reference point
        hard code to 200Ry at the moment
        """
        cutoff_pair = self.inputs.parameters.ref_cutoff_pair.get_list()
        ecutwfc = cutoff_pair[0]
        ecutrho = cutoff_pair[1]
        inputs = self.get_create_process_inputs()

        inputs['parameters']['ecutwfc'] = orm.Float(ecutwfc)
        inputs['parameters']['ecutrho'] = orm.Float(ecutrho)

        running = self.submit(self.get_create_process(), **inputs)

        self.report(
            f'launching reference CohesiveEnergyWorkChain<{running.pk}>.')

        return ToContext(ref_workchain=running)

    @abstractmethod
    def get_create_process_inputs(self) -> ty.Dict[str, ty.Any]:
        """
        The inputs used to running the create workflow
        Normally the corresponding property evaluation workflow.
        """

    @abstractmethod
    def get_evaluate_process_inputs(self) -> ty.Dict[str, ty.Any]:
        """
        The inputs used to running the evaluate workflow
        Including inputs not passed from the previous create workflow
        Normally the reference value of `run_ref` step.
        """

    @abstractmethod
    def get_output_input_mapping(self) -> orm.Dict:
        """
        return output_input_mapping from create workflow to
        evaluate workflow
        """

    def run_all(self):
        """
        Running the calculation for other points
        """
        ref_workchain = self.ctx.ref_workchain

        if not ref_workchain.is_finished_ok:
            self.report(
                f'Reference run of CohesiveEnergyWorkChain failed with exit status {ref_workchain.exit_status}'
            )
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                pk=ref_workchain.pk)

        create_evaluate_inputs = {
            'create_process': get_fullname(self.get_create_process()),
            'evaluate_process': get_fullname(self.get_evaluate_process()),
            'create': self.get_create_process_inputs(),
            'evaluate': self.get_evaluate_process_inputs(),
            'output_input_mapping': self.get_output_input_mapping(),
        }

        input_values = list(zip(self.ctx.ecutwfc_list, self.ctx.ecutrho_list))

        self.ctx.converge_y_name, self.ctx.converge_y_unit = self.get_converge_y(
        )
        inputs = {
            'engine':
            TwoInputsTwoFactorsConvergence,
            'engine_kwargs':
            orm.Dict(
                dict={
                    'input_values': input_values,
                    'tol': self._TOLERANCE,
                    'conv_thr': self._CONV_THR_CONV,
                    'input_key': 'create.parameters.ecutwfc',
                    'extra_input_key': 'create.parameters.ecutrho',
                    'result_key':
                    f'evaluate.result:{self.ctx.converge_y_name}',
                    'convergence_window': self._CONV_WINDOW
                }),
            'evaluate_process':
            CreateEvaluateWorkChain,
            'evaluate':
            create_evaluate_inputs,
        }

        running = self.submit(OptimizationWorkChain, **inputs)
        self.report(
            f'submitting cohesive energy convergence workflow pk={running.pk}.'
        )
        return ToContext(convergence_workchain=running)

    def results(self):
        """
        doc
        """
        workchain = self.ctx.convergence_workchain

        if workchain.is_finished_ok:
            self.report(f'Convergence workflow pk={workchain.pk} finish ok.')
            optimal_process_uuid = workchain.outputs.optimal_process_uuid.value
            optimal_process = orm.load_node(optimal_process_uuid)
            res = {
                'converge_ecutwfc':
                optimal_process.inputs.create__parameters__ecutwfc,
                'converge_ecutrho':
                optimal_process.inputs.create__parameters__ecutrho,
                'converge_value':
                workchain.outputs.optimal_process_output.value,
                'converge_process_uuid':
                workchain.outputs.optimal_process_uuid.value,
            }
            self.out('output_convergence_parameters',
                     orm.Dict(dict=res).store())

        if not workchain.is_finished:
            self.report(
                f'Reference run of Convergence optimize workchain pk={workchain.pk} '
                f'failed with exit status {workchain.exit_status}')
            return self.exit_codes.ERROR_SUB_PROCESS_FAILED.format(
                pk=workchain.pk)

        if workchain.is_finished and workchain.exit_status == 202:
            self.report(
                f'Convergence workflow pk={workchain.pk} is finished but not converge.'
            )

        import numpy as np

        self.ctx.children = [
            child for child in workchain.get_outgoing().all_nodes()
            if isinstance(child, orm.WorkChainNode)
        ]
        pks = [child.pk for child in self.ctx.children if child.is_finished_ok]
        if len(pks) / len(self.ctx.children) < 0.8:
            self.report(
                f'Not enough finish ok evaluate workflow, '
                f'expect 80% only get number of {len(pks)} workflow, '
                f'that is {len(pks) / len(self.ctx.children)} of total.')
            return self.exit_codes.ERROR_NOT_ENOUGH_EVALUATE_WORKFLOW.format(
                n=len(pks))

        success_child = [
            child for child in self.ctx.children if child.is_finished_ok
        ]
        ecutwfc_list = []
        ecutrho_list = []

        results_parser_dict = self.get_parsed_results()
        output_parameters_dict = {}
        for key in results_parser_dict:
            output_parameters_dict[key] = []

        for child in success_child:
            ecutwfc = child.inputs.create__parameters__ecutwfc.value
            ecutrho = child.inputs.create__parameters__ecutrho.value

            ecutwfc_list.append(ecutwfc)
            ecutrho_list.append(ecutrho)

            # loop over all output properties of evaluate process
            # set the results in the list of corresponding name.
            for key in output_parameters_dict:
                res = child.outputs.evaluate__result[key]
                output_parameters_dict[key].append(res)

        for key in list(output_parameters_dict):
            output_parameters_dict[f'{key}_description'] = results_parser_dict[
                key]

        output_parameters_dict['ecutwfc_list'] = ecutwfc_list
        output_parameters_dict['ecutrho_list'] = ecutrho_list

        xy_data_ecutwfc = orm.XyData()
        xy_data_ecutwfc.set_x(np.array(ecutwfc_list), 'wavefunction cutoff',
                              'Rydberg')

        xy_data_ecutrho = orm.XyData()
        xy_data_ecutrho.set_x(np.array(ecutrho_list), 'charge density cutoff',
                              'Rydberg')

        ys_list = []
        ys_description = []
        ys_unit = []
        for y_key, y_value in results_parser_dict.items():
            ys_list.append(np.array(output_parameters_dict[y_key]))
            ys_description.append(y_value[0])
            ys_unit.append(y_value[1])
        xy_data_ecutwfc.set_y(ys_list, ys_description, ys_unit)
        self.out('xy_data_ecutwfc', xy_data_ecutwfc.store())
        xy_data_ecutrho.set_y(ys_list, ys_description, ys_unit)
        self.out('xy_data_ecutrho', xy_data_ecutrho.store())

        output_parameters = orm.Dict(dict=output_parameters_dict)
        self.out('output_parameters', output_parameters.store())



# -*- coding: utf-8 -*-
"""
Test of process functions
"""
import pytest

from aiida import orm
from aiida.plugins import CalculationFactory


def test_birch_murnaghan_fit(data_regression):
    """Birch–Murnaghan fit test"""
    birch_murnaghan_fit = CalculationFactory(
        'sssp_workflow.birch_murnaghan_fit')

    inputs = orm.Dict(
        dict={
            'energies': {
                '0': -155.35513331955,
                '1': -155.36820553211,
                '2': -155.37566424034,
                '3': -155.37807292397,
                '4': -155.37593723854,
                '5': -155.36971834897,
                '6': -155.35983306562
            },
            'energy_unit': 'eV/atom',
            'volume_unit': 'A^3/atom',
            'volumes': {
                '0': 19.219217135197,
                '1': 19.628136649358,
                '2': 20.037056161747,
                '3': 20.445975677205,
                '4': 20.854895189195,
                '5': 21.263814702423,
                '6': 21.672734216892
            }
        })

    output_bmf = birch_murnaghan_fit(inputs)
    assert isinstance(output_bmf, orm.Dict)
    data_regression.check(output_bmf.get_dict())


def test_calculate_delta():
    """test calcfunction calculate_delta"""
    calculate_delta = CalculationFactory('sssp_workflow.calculate_delta')

    inputs = {
        'element': orm.Str('Si'),
        'V0': orm.Float(20.4530),
        'B0': orm.Float(88.545),
        'B1': orm.Float(4.31),
    }  # Si line of the file so that the delta will be exactly zero
    res = calculate_delta(**inputs)
    assert res['delta'] == 0.0


def test_calculate_delta_H():  # pylint: disable=invalid-name
    """test calcfunction calculate_delta
    make sure space in string is striped."""
    calculate_delta = CalculationFactory('sssp_workflow.calculate_delta')

    inputs = {
        'element': orm.Str('H'),
        'V0': orm.Float(17.3883),
        'B0': orm.Float(10.284),
        'B1': orm.Float(2.71),
    }  # Si line of the file so that the delta will be exactly zero
    res = calculate_delta(**inputs)
    assert res['delta'] == 0.0


def test_calculate_delta_rare_earth():
    """test calcfunction calculate_delta"""
    calculate_delta = CalculationFactory('sssp_workflow.calculate_delta')

    inputs = {
        'element': orm.Str('La'),
        'V0': orm.Float(18.77799),
        'B0': orm.Float(122.037),
        'B1': orm.Float(4.461),
    }  # LaN line of the file so that the delta will be exactly zero
    res = calculate_delta(**inputs)
    assert res['delta'] == 0.0


def test_get_v0_b0_b1():
    """
    doc
    """
    from aiida_sssp_workflow.helpers import helper_get_v0_b0_b1

    V0, B0, B1 = helper_get_v0_b0_b1('Si')
    assert V0 == 20.4530
    assert B0 == 88.545
    assert B1 == 4.31

    V0, B0, B1 = helper_get_v0_b0_b1('La')
    assert V0 == 18.77799
    assert B0 == 122.037
    assert B1 == 4.461

    V0, B0, B1 = helper_get_v0_b0_b1('F')
    assert V0 == 19.3583
    assert B0 == 74.0411
    assert B1 == 4.1599


@pytest.mark.skip(reason='Tackle after all convergence wf fixed')
def test_get_volume_from_pressure_birch_murnaghan():
    """
    doc
    """
    from aiida_sssp_workflow.workflows.convergence.pressure import helper_get_volume_from_pressure_birch_murnaghan

    V0 = 20.4530
    B0 = 88.545
    B1 = 4.31
    P = 0.01

    ret = helper_get_volume_from_pressure_birch_murnaghan(P, V0, B0, B1)

    # check that very small residual pressure correspond a small volume change
    assert ret - 20.4530 < 0.1


@pytest.mark.skip(reason='Tackle after all convergence wf fixed')
def test_phonon_frequencies_diff():
    """test of helper_get_relative_phonon_frequencies"""
    from aiida_sssp_workflow.workflows.convergence.phonon_frequencies import helper_phonon_frequencies_difference

    input_parameters = {'dynamical_matrix_0': {'frequencies': [1., 1., 1.]}}
    ref_parameters = {'dynamical_matrix_0': {'frequencies': [1., 1., 1.]}}

    res = helper_phonon_frequencies_difference(orm.Dict(dict=input_parameters),
                                               orm.Dict(dict=ref_parameters))

    assert res['relative_diff'] == 0.0
    assert res['relative_max_diff'] == 0.0
    assert res['absolute_diff'] == 0.0
    assert res['absolute_max_diff'] == 0.0



# -*- coding: utf-8 -*-
"""Define the setup for the `aiida-pseudo` plugin."""
try:
    import fastentrypoints  # pylint: disable=unused-import
except ImportError:
    # This should only occur when building the package, i.e. for `python setup.py sdist/bdist_wheel`
    pass


def setup_package():
    """Install the `aiida-pseudo` package."""
    import json
    from setuptools import setup, find_packages

    filename_setup_json = 'setup.json'
    filename_description = 'README.md'

    with open(filename_setup_json, 'r') as handle:
        setup_json = json.load(handle)

    with open(filename_description, 'r') as handle:
        description = handle.read()

    setup(packages=find_packages(),
          package_data={
              '': ['*'],
              'aiida_sssp_workflow': [
                  'REF/CIFs/*.cif', 'REF/CIFs_REN/*.cif', 'REF/UPFs/*.UPF',
                  'sssp_protocol.yml'
              ],
          },
          long_description=description,
          long_description_content_type='text/markdown',
          **setup_json)


if __name__ == '__main__':
    setup_package()



# -*- coding: utf-8 -*-
"""
calculate bands distance
"""
import numpy as np

from aiida import orm
from aiida.engine import calcfunction


@calcfunction
def calculate_bands_distance(bands_structure_a: orm.BandsData,
                             bands_parameters_a: orm.Dict,
                             bands_structure_b: orm.BandsData,
                             bands_parameters_b: orm.Dict, smearing: orm.Float,
                             is_metal: orm.Bool):
    """doc"""
    res = get_bands_distance(bands_structure_a, bands_structure_b,
                             bands_parameters_a, bands_parameters_b,
                             smearing.value, is_metal.value)

    return orm.Dict(
        dict={
            'eta_v': res.get('eta_v', None),
            'shift_v': res.get('shift_v', None),
            'max_diff_v': res.get('max_diff_v', None),
            'eta_10': res.get('eta_10', None),
            'shift_10': res.get('shift_10', None),
            'max_diff_10': res.get('max_diff_10', None),
            'bands_unit': 'eV',
        })


def get_homo(bands, num_electrons: int):
    """
    This function only work for insulator,
    therefore the num_electrons is even number.
    """
    assert num_electrons % 2 == 0, f'There are {num_electrons} electrons, must be metal'
    # get homo band
    band = bands[:, num_electrons // 2 - 1]
    return max(band)


def fermi_dirac(band_energy, fermi_energy, smearing):
    """
    The first argument can be an array
    """
    old_settings = np.seterr(over='raise', divide='raise')
    try:
        res = 1.0 / (np.exp((band_energy - fermi_energy) / smearing) + 1.0)
    except FloatingPointError:
        res = np.heaviside(fermi_energy - band_energy, 1.0)
    np.seterr(**old_settings)

    return res


def retrieve_bands(bandsdata: orm.BandsData, start_band, num_electrons, efermi,
                   smearing, is_metal):
    """
    docstring
    """
    from efermi import pyefermi

    bands = bandsdata.get_bands()
    bands = bands - efermi  # shift all bands to fermi energy 0
    bands = bands[:, start_band:]
    output_bands = orm.ArrayData()
    output_bands.set_array('kpoints', bandsdata.get_kpoints())
    output_bands.set_array('bands', bands)

    if is_metal:
        nelectrons = num_electrons
        nkpoints = np.shape(bands)[0]
        weights = np.ones(nkpoints) / nkpoints
        bands = np.asfortranarray(bands)
        meth = 2  # firmi-dirac smearing

        output_efermi = pyefermi(bands, weights, nelectrons, smearing,
                                 nkpoints, meth)

    else:
        homo_energy = get_homo(bands, num_electrons)
        output_efermi = homo_energy

    return {
        'bands': output_bands,
        'efermi': output_efermi,
    }


def calculate_eta_and_max_diff(bands_a: orm.ArrayData, bands_b: orm.ArrayData,
                               efermi_a, efermi_b, fermi_shift, smearing):
    """
    docstring
    """
    from functools import partial
    from scipy.optimize import minimize

    bands_a = bands_a.get_array('bands')
    bands_b = bands_b.get_array('bands')
    num_bands = min(np.shape(bands_a)[1], np.shape(bands_b)[1])

    assert np.shape(bands_a)[0] == np.shape(
        bands_b)[0], 'have different kpoints'

    # truncate the bands to same size
    bands_a = bands_a[:, :num_bands - 1]
    bands_b = bands_b[:, :num_bands - 1]

    occ_a = fermi_dirac(bands_a, efermi_a + fermi_shift, smearing)
    occ_b = fermi_dirac(bands_b, efermi_b + fermi_shift, smearing)
    occ = np.sqrt(occ_a * occ_b)

    bands_diff = bands_a - bands_b

    def fun_shift(occ, bands_diff, shift):
        return np.sqrt(np.sum(occ * (bands_diff + shift)**2) / np.sum(occ))

    # Compute eta
    eta_val = partial(fun_shift, occ, bands_diff)
    results = minimize(eta_val, np.array([0.0]), method='Nelder-Mead')
    eta = results.get('fun')
    shift = results.get('x')[0]

    # Compute max_diff:
    # then find from abs_diff the max one of which the occ > 0.5
    # first make occ > 0.5 to be 1 and occ <= 0.5 to be 0, then element-wise the
    # new occ matrix with abs_diff, and find the max diff
    abs_diff = np.abs(bands_diff + shift)
    new_occ = np.heaviside(occ - 0.5, 0.0)
    new_diff = np.multiply(abs_diff, new_occ)
    max_diff = np.amax(new_diff)

    return {
        'eta': eta,
        'shift': shift,
        'max_diff': max_diff,
    }


def get_bands_distance(bands_a: orm.BandsData, bands_b: orm.BandsData,
                       band_parameters_a: orm.Dict,
                       band_parameters_b: orm.Dict, smearing, is_metal):
    """
    TODO docstring
    """
    num_electrons_a = band_parameters_a['number_of_electrons']
    num_electrons_b = band_parameters_b['number_of_electrons']
    efermi_a = band_parameters_a['fermi_energy']
    efermi_b = band_parameters_b['fermi_energy']

    if num_electrons_a <= num_electrons_b:
        num_electrons = int(num_electrons_a)
        res = retrieve_bands(bands_a, 0, num_electrons, efermi_a, smearing,
                             is_metal)
        bands_a = res['bands']
        efermi_a = res['efermi']

        start_band = int(num_electrons_b - num_electrons_a) // 2
        res = retrieve_bands(bands_b, start_band, num_electrons, efermi_b,
                             smearing, is_metal)
        bands_b = res['bands']
        efermi_b = res['efermi']
    else:
        # num_electrons_b < num_electrons_a:
        num_electrons = int(num_electrons_b)
        start_band = int(num_electrons_a - num_electrons_b) // 2
        res = retrieve_bands(bands_a, start_band, num_electrons, efermi_a,
                             smearing, is_metal)
        bands_a = res['bands']
        efermi_a = res['efermi']

        res = retrieve_bands(bands_b, 0, num_electrons, efermi_b, smearing,
                             is_metal)
        bands_b = res['bands']
        efermi_b = res['efermi']

    # eta_v
    fermi_shift = 0.0
    if is_metal:
        smearing_v = smearing
    else:
        smearing_v = 0

    outputs = calculate_eta_and_max_diff(bands_a, bands_b, efermi_a, efermi_b,
                                         fermi_shift, smearing_v)
    eta_v = outputs.get('eta')
    shift_v = outputs.get('shift')
    max_diff_v = outputs.get('max_diff')

    # eta_10
    fermi_shift = 10.0
    # if not metal
    smearing_10 = smearing
    outputs = calculate_eta_and_max_diff(bands_a, bands_b, efermi_a, efermi_b,
                                         fermi_shift, smearing_10)

    eta_10 = outputs.get('eta')
    shift_10 = outputs.get('shift')
    max_diff_10 = outputs.get('max_diff')

    return {
        'eta_v': eta_v,
        'shift_v': shift_v,
        'max_diff_v': max_diff_v,
        'eta_10': eta_10,
        'shift_10': shift_10,
        'max_diff_10': max_diff_10,
    }



# -*- coding: utf-8 -*-
"""tests of customize optimize engine"""
from aiida import orm
from aiida.plugins import WorkflowFactory
from aiida.engine import run
from aiida_sssp_workflow.workflows.convergence.engine import TwoInputsTwoFactorsConvergence

from .sample_processes import echo_calcfunction

OptimizationWorkChain = WorkflowFactory('optimize.optimize')

_INPUT_VALUES = [(1, 2), (1, 1), (1.0, 0.3), (1.0, 0.2), (0.002, 0.100),
                 (0.0793, 0.02), (0.0794, 0.02), (0.0792, 0.02)]


def test_condition_1():
    """
    Condition 1:
    conv_thr satisfied first and convergence depend on tol < 1e-2
    result: 0.102
    """
    inputs = {
        'engine':
        TwoInputsTwoFactorsConvergence,
        'engine_kwargs':
        orm.Dict(
            dict={
                'input_values': _INPUT_VALUES,
                'tol': 1e-2,
                'conv_thr': 2,
                'input_key': 'x',
                'extra_input_key': 'y',
                'result_key': 'result',
                'convergence_window': 2
            }),
        'evaluate_process':
        echo_calcfunction
    }

    res = run(OptimizationWorkChain, **inputs)
    assert res['optimal_process_output'] == 0.102


def test_condition_2():
    """
    Condition 2:
    tol satisfied first and convergence depend on conv_thr < 1e-1
    result: 0.0993
    """
    inputs = {
        'engine':
        TwoInputsTwoFactorsConvergence,
        'engine_kwargs':
        orm.Dict(
            dict={
                'input_values': _INPUT_VALUES,
                'tol': 1e-2,
                'conv_thr': 1e-1,
                'input_key': 'x',
                'extra_input_key': 'y',
                'result_key': 'result',
                'convergence_window': 2
            }),
        'evaluate_process':
        echo_calcfunction
    }

    res = run(OptimizationWorkChain, **inputs)
    assert res['optimal_process_output'] == 0.0993



# -*- coding: utf-8 -*-
"""
Bands distance of many input pseudos
"""
import yaml
import importlib_resources

from aiida import orm
from aiida.engine import WorkChain, if_, append_
from aiida.plugins import DataFactory

from aiida_sssp_workflow.utils import NONMETAL_ELEMENTS, update_dict, \
    RARE_EARTH_ELEMENTS, \
    get_standard_cif_filename_from_element
from aiida_sssp_workflow.workflows.evaluate._bands import BandsWorkChain
from aiida_sssp_workflow.calculations.calculate_bands_distance import calculate_bands_distance

UpfData = DataFactory('pseudo.upf')


def validate_input_pseudos(d_pseudos, _):
    """Validate that all input pseudos map to same element"""
    element = set(pseudo.element for pseudo in d_pseudos.values())

    if len(element) > 1:
        return f'The pseudos corespond to different elements {element}.'


class BandsDistanceWorkChain(WorkChain):
    """WorkChain to do bands comparision of different pseudos"""

    _MAX_WALLCLOCK_SECONDS = 1800 * 3

    _LARGE_DUAL_ELEMENTS = ['Fe', 'Hf']

    _INIT_NBANDS_FACTOR = 3.0
    _RY_TO_EV = 13.6056980659

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        super().define(spec)
        # yapf: disable
        spec.input('code', valid_type=orm.Code,
                    help='The `pw.x` code use for the `PwCalculation`.')
        spec.input_namespace('input_pseudos', valid_type=UpfData, dynamic=True,
                    validator=validate_input_pseudos,
                    help='A mapping of `UpfData` node to be verified onto file name:upf.')
        spec.input('protocol', valid_type=orm.Str, default=lambda: orm.Str('efficiency'),
                    help='The protocol to use for the workchain.')
        spec.input('options', valid_type=orm.Dict, required=False,
                    help='Optional `options` to use for the `PwCalculations`.')
        spec.input('parallelization', valid_type=orm.Dict, required=False,
                    help='Parallelization options for the `PwCalculations`.')
        spec.input('clean_workdir', valid_type=orm.Bool, default=lambda: orm.Bool(False),
                    help='If `True`, work directories of all called calculation will be cleaned at the end of execution.')

        spec.outline(
            cls.init_setup,
            if_(cls.is_rare_earth_element)(
                cls.extra_setup_for_rare_earth_element, ),
            cls.setup_code_parameters_from_protocol,
            cls.setup_code_resource_options,
            cls.run_bands_evaluation,
            cls.calculate_bands_distance,
        )

        spec.output('output_bands_distance', valid_type=orm.Dict, required=True,
            help='the distance of each pair of pseudopotentials.')

    def _get_protocol(self):
        """Load and read protocol from faml file to a verbose dict"""
        import_path = importlib_resources.path('aiida_sssp_workflow',
                                               'sssp_protocol.yml')
        with import_path as pp_path, open(pp_path, 'rb') as handle:
            self._protocol = yaml.safe_load(handle)  # pylint: disable=attribute-defined-outside-init

            return self._protocol

    def init_setup(self):
        """
        This step contains all preparation before actaul setup, e.g. set
        the context of element, base_structure, base pw_parameters and pseudos.
        """
        # parse pseudo and output its header information
        self.ctx.pw_parameters = {}
        self.ctx.extra_parameters = {}

        # First idx since already being validate to make sure all mapping to same element
        element = [pseudo.element for pseudo in self.inputs.input_pseudos.values()][0]
        self.ctx.element = element

        # Structures for convergence verification are all primitive structures
        # the original conventional structure comes from the same CIF files of
        # http:// molmod.ugent.be/deltacodesdft/
        # EXCEPT that for the element fluorine the `SiF4.cif` used for convergence
        # reason. But we do the structure setup for SiF4 in the following step:
        # `cls.extra_setup_for_fluorine_element`
        cif_file = get_standard_cif_filename_from_element(element)
        self.ctx.structure = orm.CifData.get_or_create(
            cif_file)[0].get_structure(primitive_cell=True)

        # extra setting for bands convergence
        self.ctx.is_metal = element not in NONMETAL_ELEMENTS

    def is_rare_earth_element(self):
        """Check if the element is rare earth"""
        return self.ctx.element in RARE_EARTH_ELEMENTS

    def extra_setup_for_rare_earth_element(self):
        """Extra setup for rare earth element"""
        import_path = importlib_resources.path('aiida_sssp_workflow.REF.UPFs',
                                               'N.pbe-n-radius_5.UPF')
        with import_path as pp_path, open(pp_path, 'rb') as stream:
            upf_nitrogen = UpfData(stream)
            self.ctx.pseudo_N = upf_nitrogen

        # In rare earth case, increase the initial number of bands,
        # otherwise the occupation will not fill up in the highest band
        # which always trigger the `PwBaseWorkChain` sanity check.
        nbands = self.inputs.pseudo.z_valence + upf_nitrogen.z_valence // 2
        nbands_factor = 2

        self.ctx.extra_parameters = {
            'SYSTEM': {
                'nbnd': int(nbands * nbands_factor),
            },
        }

    def setup_code_parameters_from_protocol(self):
        """Input validation"""
        # pylint: disable=invalid-name, attribute-defined-outside-init

        # Read from protocol if parameters not set from inputs
        protocol_name = self.inputs.protocol.value
        protocol = self._get_protocol()[protocol_name]
        protocol = protocol['bands_distance']
        self._DEGAUSS = protocol['degauss']
        self._OCCUPATIONS = protocol['occupations']
        self._SMEARING = protocol['smearing']
        self._CONV_THR = protocol['electron_conv_thr']
        self._KDISTANCE = protocol['kpoints_distance']

        self.ctx.ecutwfc = protocol['ecutwfc']
        self.ctx.kpoints_distance = self._KDISTANCE
        self.ctx.degauss = self._DEGAUSS

        # set large ecutrho for specific elements
        if self.ctx.element in self._LARGE_DUAL_ELEMENTS:
            self.ctx.ecutrho = protocol['ecutrho'] * 2
        else:
            self.ctx.ecutrho = protocol['ecutrho']

        self.ctx.pw_parameters = {
            'SYSTEM': {
                'degauss': self._DEGAUSS,
                'occupations': self._OCCUPATIONS,
                'smearing': self._SMEARING,
            },
            'ELECTRONS': {
                'conv_thr': self._CONV_THR,
            },
        }

        self.ctx.pw_parameters = update_dict(self.ctx.pw_parameters,
                                        self.ctx.extra_parameters)

        self.report(
            f'The pw parameters for convergence is: {self.ctx.pw_parameters}'
        )

    def setup_code_resource_options(self):
        """
        setup resource options and parallelization for `PwCalculation` from inputs
        """
        if 'options' in self.inputs:
            self.ctx.options = self.inputs.options.get_dict()
        else:
            from aiida_sssp_workflow.utils import get_default_options

            self.ctx.options = get_default_options(
                max_wallclock_seconds=self._MAX_WALLCLOCK_SECONDS,
                with_mpi=True)

        if 'parallelization' in self.inputs:
            self.ctx.parallelization = self.inputs.parallelization.get_dict()
        else:
            self.ctx.parallelization = {}

        self.report(f'resource options set to {self.ctx.options}')
        self.report(
            f'parallelization options set to {self.ctx.parallelization}')

    def _get_inputs(self, element, pseudos):
        """
        get inputs for the bands evaluation with given pseudo
        """
        if element in RARE_EARTH_ELEMENTS:
            pseudos['N'] = self.ctx.pseudo_N

        # yapf: disable
        inputs = {
            'code': self.inputs.code,
            'pseudos': pseudos,
            'structure': self.ctx.structure,
            'pw_base_parameters': orm.Dict(dict=self.ctx.pw_parameters),
            'ecutwfc': orm.Float(self.ctx.ecutwfc),
            'ecutrho': orm.Float(self.ctx.ecutrho),
            'kpoints_distance': orm.Float(self.ctx.kpoints_distance),
            'init_nbands_factor': orm.Float(self._INIT_NBANDS_FACTOR),
            'options': orm.Dict(dict=self.ctx.options),
            'parallelization': orm.Dict(dict=self.ctx.parallelization),
            'clean_workdir': orm.Bool(False),   # will leave the workdir clean to outer most wf
        }
        # yapf: enable

        return inputs

    def run_bands_evaluation(self):
        """run bands evaluation of pp in inputs list"""
        for name, pseudo in self.inputs.input_pseudos.items():
            pseudos = {self.ctx.element: pseudo}
            inputs = self._get_inputs(self.ctx.element, pseudos)

            running = self.submit(BandsWorkChain, **inputs)

            # set the name to indicate the pp
            running.set_extra('name', name)
            self.report(
                f'launching pseudo >{name}<: {pseudo} BandsWorkChain<{running.pk}>'
            )

            self.to_context(children=append_(running))

    def calculate_bands_distance(self):
        """calculate bands distance of every pair of pp"""
        from itertools import combinations

        children = self.ctx.children
        success_children = [
            child for child in children if child.is_finished_ok
        ]

        d_output_parameters = {}

        for idx, (wca, wcb) in enumerate(combinations(success_children, 2)):
            d_ab = {
                'workchain_name_a': wca.extras['name'],
                'workchain_name_b': wcb.extras['name']
            }

            wca_bands_parameters = wca.outputs.output_bands_parameters
            wcb_bands_parameters = wcb.outputs.output_bands_parameters
            wca_bands_structure = wca.outputs.output_bands_structure
            wcb_bands_structure = wcb.outputs.output_bands_structure

            res = calculate_bands_distance(
                bands_structure_a=wca_bands_structure,
                bands_parameters_a=wca_bands_parameters,
                bands_structure_b=wcb_bands_structure,
                bands_parameters_b=wcb_bands_parameters,
                smearing=orm.Float(self.ctx.degauss * self._RY_TO_EV),
                is_metal=orm.Bool(self.ctx.is_metal))

            d_ab.update(res.get_dict())
            d_output_parameters[f'dist_{idx}'] = d_ab

        self.out('output_bands_distance',
                 orm.Dict(dict=d_output_parameters).store())
